@@ -1,7 +1,7 @@
 /**
  * ntfs-3g_common.c - Common definitions for ntfs-3g and lowntfs-3g.
  *
- * Copyright (c) 2010-2011 Jean-Pierre Andre
+ * Copyright (c) 2010-2019 Jean-Pierre Andre
  * Copyright (c) 2010      Erik Larsson
  *
  * This program/include file is free software; you can redistribute it and/or
@@ -28,6 +28,10 @@
 #include <stdlib.h>
 #endif
 
+#ifdef HAVE_DLFCN_H
+#include <dlfcn.h>
+#endif
+
 #ifdef HAVE_STRING_H
 #include <string.h>
 #endif
@@ -43,9 +47,13 @@
 #include <getopt.h>
 #include <fuse.h>
 
+#include "compat.h"
 #include "inode.h"
+#include "dir.h"
 #include "security.h"
 #include "xattrs.h"
+#include "reparse.h"
+#include "plugin.h"
 #include "ntfs-3g_common.h"
 #include "realpath.h"
 #include "misc.h"
@@ -76,6 +84,8 @@ const struct DEFOPTION optionlist[] = {
 	{ "noatime", OPT_NOATIME, FLGOPT_BOGUS },
 	{ "atime", OPT_ATIME, FLGOPT_BOGUS },
 	{ "relatime", OPT_RELATIME, FLGOPT_BOGUS },
+	{ "delay_mtime", OPT_DMTIME, FLGOPT_DECIMAL | FLGOPT_OPTIONAL },
+	{ "rw", OPT_RW, FLGOPT_BOGUS },
 	{ "fake_rw", OPT_FAKE_RW, FLGOPT_BOGUS },
 	{ "fsname", OPT_FSNAME, FLGOPT_NOSUPPORT },
 	{ "no_def_opts", OPT_NO_DEF_OPTS, FLGOPT_BOGUS },
@@ -154,6 +164,54 @@ int ntfs_strappend(char **dest, const char *append)
 	return 0;
 }
 
+/*
+ *		Insert an option before ",fsname="
+ *	This is for keeping "fsname" as the last option, because on
+ *	Solaris device names may contain commas.
+ */
+
+int ntfs_strinsert(char **dest, const char *append)
+{
+	char *p, *q;
+	size_t size_append, size_dest = 0;
+	
+	if (!dest)
+		return -1;
+	if (!append)
+		return 0;
+
+	size_append = strlen(append);
+	if (*dest)
+		size_dest = strlen(*dest);
+	
+	if (strappend_is_large(size_dest) || strappend_is_large(size_append)) {
+		errno = EOVERFLOW;
+		ntfs_log_perror("%s: Too large input buffer", EXEC_NAME);
+		return -1;
+	}
+	
+	p = (char*)malloc(size_dest + size_append + 1);
+	if (!p) {
+		ntfs_log_perror("%s: Memory reallocation failed", EXEC_NAME);
+		return -1;
+	}
+	strcpy(p, *dest);
+	q = strstr(p, ",fsname=");
+	if (q) {
+		strcpy(q, append);
+		q = strstr(*dest, ",fsname=");
+		if (q)
+			strcat(p, q);
+		free(*dest);
+		*dest = p;
+	} else {
+		free(*dest);
+		*dest = p;
+		strcpy(*dest + size_dest, append);
+	}
+	return 0;
+}
+
 static int bogus_option_value(char *val, const char *s)
 {
 	if (val) {
@@ -213,12 +271,17 @@ char *parse_mount_options(ntfs_fuse_context_t *ctx,
 					opt);
 				goto err_exit;
 			}
-			if ((poptl->flags & FLGOPT_DECIMAL)
-			    && (!val
-				|| !sscanf(val, "%i", &intarg))) {
-				ntfs_log_error("'%s' option needs a decimal value\n",
-					opt);
-				goto err_exit;
+			if (poptl->flags & FLGOPT_DECIMAL) {
+				if ((poptl->flags & FLGOPT_OPTIONAL) && !val)
+					intarg = 0;
+				else
+					if (!val
+					    || !sscanf(val, "%i", &intarg)) {
+						ntfs_log_error("'%s' option "
+						     "needs a decimal value\n",
+							opt);
+						goto err_exit;
+					}
 			}
 			if ((poptl->flags & FLGOPT_STRING)
 			    && missing_option_value(val, opt))
@@ -229,6 +292,9 @@ char *parse_mount_options(ntfs_fuse_context_t *ctx,
 			case OPT_FAKE_RW :
 				ctx->ro = TRUE;
 				break;
+			case OPT_RW :
+				ctx->rw = TRUE;
+				break;
 			case OPT_NOATIME :
 				ctx->atime = ATIME_DISABLED;
 				break;
@@ -238,6 +304,11 @@ char *parse_mount_options(ntfs_fuse_context_t *ctx,
 			case OPT_RELATIME :
 				ctx->atime = ATIME_RELATIVE;
 				break;
+			case OPT_DMTIME :
+				if (!intarg)
+					intarg = DEFAULT_DMTIME;
+				ctx->dmtime = intarg*10000000LL;
+				break;
 			case OPT_NO_DEF_OPTS :
 				no_def_opts = TRUE; /* Don't add default options. */
 				ctx->silent = FALSE; /* cancel default silent */
@@ -472,8 +543,11 @@ char *parse_mount_options(ntfs_fuse_context_t *ctx,
 		ctx->secure_flags |= (1 << SECURITY_ACL);
 	if (want_permissions)
 		ctx->secure_flags |= (1 << SECURITY_WANTED);
-	if (ctx->ro)
+	if (ctx->ro) {
 		ctx->secure_flags &= ~(1 << SECURITY_ADDSECURIDS);
+		ctx->hiberfile = FALSE;
+		ctx->rw = FALSE;
+	}
 exit:
 	free(options);
 	return ret;
@@ -494,7 +568,7 @@ int ntfs_parse_options(struct ntfs_options *popts, void (*usage)(void),
 {
 	int c;
 
-	static const char *sopt = "-o:hnvV";
+	static const char *sopt = "-o:hnsvV";
 	static const struct option lopt[] = {
 		{ "options",	 required_argument,	NULL, 'o' },
 		{ "help",	 no_argument,		NULL, 'h' },
@@ -548,6 +622,11 @@ int ntfs_parse_options(struct ntfs_options *popts, void (*usage)(void),
 			 * no effect - automount passes it, meaning 'no-mtab'
 			 */
 			break;
+		case 's':
+			/*
+			 * no effect - automount passes it, meaning sloppy
+			 */
+			break;
 		case 'v':
 			/*
 			 * We must handle the 'verbose' option even if
@@ -684,3 +763,171 @@ exit :
 }
 
 #endif /* HAVE_SETXATTR */
+
+#ifndef DISABLE_PLUGINS
+
+int register_reparse_plugin(ntfs_fuse_context_t *ctx, le32 tag,
+				const plugin_operations_t *ops, void *handle)
+{
+	plugin_list_t *plugin;
+	int res;
+
+	res = -1;
+	plugin = (plugin_list_t*)ntfs_malloc(sizeof(plugin_list_t));
+	if (plugin) {
+		plugin->tag = tag;
+		plugin->ops = ops;
+		plugin->handle = handle;
+		plugin->next = ctx->plugins;
+		ctx->plugins = plugin;
+		res = 0;
+	}
+	return (res);
+}
+
+/*
+ *		Get the reparse operations associated to an inode
+ *
+ *	The plugin able to process the reparse point is dynamically loaded
+ *
+ *	When successful, returns the operations vector and the reparse
+ *		data if requested,
+ *	Otherwise returns NULL, with errno set.
+ */
+
+const struct plugin_operations *select_reparse_plugin(ntfs_fuse_context_t *ctx,
+				ntfs_inode *ni, REPARSE_POINT **reparse_wanted)
+{
+	const struct plugin_operations *ops;
+	void *handle;
+	REPARSE_POINT *reparse;
+	le32 tag, seltag;
+	plugin_list_t *plugin;
+	plugin_init_t pinit;
+
+	ops = (struct plugin_operations*)NULL;
+	reparse = ntfs_get_reparse_point(ni);
+	if (reparse) {
+		tag = reparse->reparse_tag;
+		seltag = tag & IO_REPARSE_PLUGIN_SELECT;
+		for (plugin=ctx->plugins; plugin && (plugin->tag != seltag);
+						plugin = plugin->next) { }
+		if (plugin) {
+			ops = plugin->ops;
+		} else {
+#ifdef PLUGIN_DIR
+			char name[sizeof(PLUGIN_DIR) + 64];
+
+			snprintf(name,sizeof(name), PLUGIN_DIR
+					"/ntfs-plugin-%08lx.so",
+					(long)le32_to_cpu(seltag));
+#else
+			char name[64];
+
+			snprintf(name,sizeof(name), "ntfs-plugin-%08lx.so",
+					(long)le32_to_cpu(seltag));
+#endif
+			handle = dlopen(name, RTLD_LAZY);
+			if (handle) {
+				pinit = (plugin_init_t)dlsym(handle, "init");
+				if (pinit) {
+				/* pinit() should set errno if it fails */
+					ops = (*pinit)(tag);
+					if (ops && register_reparse_plugin(ctx,
+							seltag, ops, handle))
+						ops = (struct plugin_operations*)NULL;
+				} else
+					errno = ELIBBAD;
+				if (!ops)
+					dlclose(handle);
+			} else {
+				errno = ELIBACC;
+				if (!(ctx->errors_logged & ERR_PLUGIN)) {
+					ntfs_log_perror(
+						"Could not load plugin %s",
+						name);
+					ntfs_log_error("Hint %s\n",dlerror());
+				}
+				ctx->errors_logged |= ERR_PLUGIN;
+			}
+		}
+		if (ops && reparse_wanted)
+			*reparse_wanted = reparse;
+		else
+			free(reparse);
+	}
+	return (ops);
+}
+
+void close_reparse_plugins(ntfs_fuse_context_t *ctx)
+{
+	while (ctx->plugins) {
+		plugin_list_t *next;
+
+		next = ctx->plugins->next;
+		if (ctx->plugins->handle)
+			dlclose(ctx->plugins->handle);
+		free(ctx->plugins);
+		ctx->plugins = next;
+	}
+}
+
+#endif /* DISABLE_PLUGINS */
+
+#ifdef HAVE_SETXATTR
+
+/*
+ *		Check whether a user xattr is allowed
+ *
+ *	The inode must be a plain file or a directory. The only allowed
+ *	metadata file is the root directory (useful for MacOSX and hopefully
+ *	does not harm Windows).
+ */
+
+BOOL user_xattrs_allowed(ntfs_fuse_context_t *ctx __attribute__((unused)),
+			ntfs_inode *ni)
+{
+	u32 dt_type;
+	BOOL res;
+
+		/* Quick return for common cases and root */
+	if (!(ni->flags & (FILE_ATTR_SYSTEM | FILE_ATTR_REPARSE_POINT))
+	    || (ni->mft_no == FILE_root))
+		res = TRUE;
+	else {
+			/* Reparse point depends on kind, see plugin */
+		if (ni->flags & FILE_ATTR_REPARSE_POINT) {
+#ifndef DISABLE_PLUGINS
+			struct stat stbuf;
+			REPARSE_POINT *reparse;
+			const plugin_operations_t *ops;
+
+			res = FALSE; /* default for error cases */
+			ops = select_reparse_plugin(ctx, ni, &reparse);
+			if (ops) {
+				if (ops->getattr
+				    && !ops->getattr(ni,reparse,&stbuf)) {
+					res = S_ISREG(stbuf.st_mode)
+						    || S_ISDIR(stbuf.st_mode);
+				}
+				free(reparse);
+			}
+#else /* DISABLE_PLUGINS */
+			res = FALSE; /* mountpoints, symlinks, ... */
+#endif /* DISABLE_PLUGINS */
+		} else {
+				/* Metadata */
+			if (ni->mft_no < FILE_first_user)
+				res = FALSE;
+			else {
+				/* Interix types */
+				dt_type = ntfs_interix_types(ni);
+				res = (dt_type == NTFS_DT_REG)
+					|| (dt_type == NTFS_DT_DIR);
+			}
+		}
+	}
+	return (res);
+}
+
+#endif /* HAVE_SETXATTR */
